from sentenai.stream.events import Event
import pytest, time
from datetime import datetime, timedelta


@pytest.fixture(scope='module')
def test_stream(client, request):
    if client.streams['test-stats']:
        del client.streams['test-stats']
    client.streams['test-stats'].init()
    time.sleep(1)
    def teardown():
        del client.streams['test-stats']
    for i in range(20):
        client.streams['test-stats'].events.insert(Event(id=str(i), ts=datetime(2010,1,1,0,0,i), data={"i": i}))
    time.sleep(10)
    request.addfinalizer(teardown)
    return client.streams['test-stats']

def test_field_value(client, test_stream):
    x = test_stream['i'].value()
    assert x == 19

def test_field_value_at(client, test_stream):
    ts = test_stream.events[5].ts
    x = test_stream['i'].value(at=ts)
    assert x == 5

def test_field_value_at_operator(client, test_stream):
    ts = test_stream.events[5].ts
    assert test_stream['i'].value(at=ts) == test_stream['i'] @ ts

def test_mean(client, test_stream):
    assert test_stream['i'].mean == 9.5

def test_min(client, test_stream):
    assert test_stream['i'].min == 0

def test_max(client, test_stream):
    assert test_stream['i'].max == 19

def test_count(client, test_stream):
    assert test_stream['i'].count == 20



from hypothesis import given, example, assume
from hypothesis.strategies import text

from sentenai.utils import py2str, PY3
import string

@given(text())
@example(string.ascii_letters)
def test_py2str(s):

    @py2str
    class Foo:
        def __init__(self, s):
            self.s = s

        def __str__(self):
            return self.s

    assume(isinstance(str(Foo(s)), str))





from sentenai.stream.events import Event
from datetime import datetime
import pytest, time
from uuid import uuid4

def test_create_event(client, foo):
    e = foo.events.insert(Event(ts=datetime.utcnow(), data={'foo': 'bar'}))
    assert isinstance(e, Event)

def test_insert_event_into_nonexistent_stream(client):
    i = uuid4().hex
    t = datetime.utcnow()
    s = client.streams[i]
    with pytest.raises(Exception):
        s.events.insert(Event(ts=t, data={'foo': 'bar'}))
    if s:
        del client[i]


def test_get_event(client, foo):
    e = foo.events.insert(Event(ts=datetime.utcnow(), data={'foo': 'bar'}))
    assert e == foo.events[e.id]

def test_remove_event(client, foo):
    foo_evts = foo.events
    e = foo_evts.insert(Event(ts=datetime.utcnow(), data={'foo': 'bar'}))
    foo_evts.remove(e)
    with pytest.raises(KeyError):
        foo_evts[e.id]

def test_remove_non_existent_event(client, foo):
    e = Event(id="a", ts=datetime.utcnow(), data={'foo': 'bar'})
    with pytest.raises(KeyError):
        foo.events.remove(e)


def test_update_event(client, foo):
    foo_evts = foo.events
    e = foo_evts.insert(Event(ts=datetime.utcnow(), data={'foo': 'bar'}))
    e.data = {"bar": "baz"}
    foo_evts.update(e)
    assert foo_evts[e.id].data == {"bar": "baz"}

@pytest.fixture(scope='module')
def foo(client, request):
    x = client.streams['foo']
    if not x:
        x.init()
    return x

@pytest.fixture(scope='module')
def evts_stream(client, request):
    if client.streams['pytest-events']:
        del client.streams['pytest-events']
    client.streams['pytest-events'].init()
    time.sleep(1)
    def teardown():
        #for i in range(20):
        #    del client.streams['pytest-events'].events[str(i)]
        del client.streams['pytest-events']
    for i in range(20):
        client.streams['pytest-events'].events.insert(Event(id=str(i), ts=datetime(2015,1,1+i), data={"i": i}))
    request.addfinalizer(teardown)
    time.sleep(10)
    return client.streams['pytest-events']


def test_events_length(client, evts_stream):
    assert len(evts_stream.events) == 20

def test_events_range_int(client, evts_stream):
    x = evts_stream.events[:10]
    assert len(x) == 10

def test_events_range_int_vals(client, evts_stream):
    x = evts_stream.events[:10]
    for i in range(10):
        assert i == x[i].data['i']

def test_events_newest(client, evts_stream):
    x = evts_stream.events[-1]
    assert x.data['i'] == 19

def test_events_oldest(client, evts_stream):
    x = evts_stream.events[0]
    assert x.data['i'] == 0

def test_events_third_oldest(client, evts_stream):
    x = evts_stream.events[2]
    assert x.data['i'] == 2

def test_events_third_newest(client, evts_stream):
    x = evts_stream.events[-3]
    assert x.data['i'] == 17

def test_events_range_datetimeleft(client, evts_stream):
    x = evts_stream.events[datetime(2015,1,1):]
    assert len(x) == 20
    for i, e in zip(range(25), x):
        assert i == x[i].data['i']

def test_events_range_datetime_right(client, evts_stream):
    x = evts_stream.events[:datetime(2015,1,5)]
    assert len(x) == 4
    for i, e in zip(range(25), x):
        assert i == x[i].data['i']

def test_events_range_datetime_full(client, evts_stream):
    x = evts_stream.events[datetime(2015,1,1):datetime(2015,1,21)]
    assert len(x) == 20
    for i, e in zip(range(25), x):
        assert i == x[i].data['i']

def test_events_range_datetime_full2(client, evts_stream):
    x = evts_stream.events[datetime(2015,1,1):datetime(2015,1,21)]
    assert len(x) == 20
    for i, e in zip(range(25), x):
        assert i == x[i].data['i']

def test_events_range_full_limit(client, evts_stream):
    x = evts_stream.events[datetime(2015,1,3):datetime(2015,1,15):5]
    assert len(x) == 5
    for i, e in zip(range(25), x):
        assert i + 2 == x[i].data['i']

def test_events_range_full(client, evts_stream):
    x = evts_stream.events[:]
    assert len(x) == 20
    for i, e in zip(range(25), x):
        assert i == x[i].data['i']



import pytest
from datetime import datetime
from sentenai.stream.fields import Field, Fields

def test_values(client):
    assert type((client.streams['weather'] @ datetime.utcnow())['humidity']) == float


def test_fields(client):
    for x in client.streams['weather']:
        assert isinstance(x, Field)

def test_field_values(client):
    assert type(client.streams['weather']['humidity'] @ datetime.utcnow()) == float

def test_field_mean(client):
    assert type(client.streams['weather']['humidity'].mean) == float

def test_field_min(client):
    assert type(client.streams['weather']['humidity'].min) == float

def test_field_max(client):
    assert type(client.streams['weather']['humidity'].max) == float

def test_field_std(client):
    assert type(client.streams['weather']['humidity'].std) == float

def test_field_unique(client):
    u = client.streams['weather']['icon'].unique
    assert type(u) == dict





from sentenai.api import API
from sentenai.stream import Streams


def test_ping(client):
    assert client.ping() > 0.

def test_streams(client):
    assert isinstance(client.streams, Streams) and isinstance(client.streams, API)




from hypothesis import given, example, assume
from hypothesis.strategies import text, tuples, uuids, one_of, none, integers, floats, datetimes

from sentenai import Sentenai
from sentenai.api.stream import Stream
from sentenai.utils import iso8601
import string, unittest, requests_mock, requests, pytest
from datetime import datetime

try:
    from urllib.parse import quote
except:
    from urllib import quote

def stream(name, *filters):
    return Stream(None, name, {}, None, True, *filters)

URL = "https://api.sentenai.com/"
URL_STREAMS   = URL + "streams"
URL_STREAM_ID = URL + "streams/{}"
URL_EVENTS    = URL + "streams/{}/events"
URL_EVENTS_ID = URL + "streams/{}/events/{}"
URL_VALUES    = URL + "streams/{}/values"
# TODO: this is changing
URL_STREAM_ID_META = URL + "streams/{}"
URL_STREAM_RANGE = URL + "streams/{}/start/{}/end/{}"

def range_url(stream, start, end):
    return URL_STREAM_RANGE.format(stream.name, iso8601(start), iso8601(end))

test_client = Sentenai(auth_key = "")

def test_streams_call():
    with requests_mock.mock() as m:
        m.get(URL_STREAMS, json=[])
        resp = test_client.streams()
        assume(list(resp) == [])

def test_stream_existence():
    with requests_mock.mock() as m:
        s = test_client.Stream('real-stream')
        m.get(URL_STREAM_ID.format(s.name), status_code=200)
        assume(bool(s) == True)

def test_stream_nonexistence():
    with requests_mock.mock() as m:
        s = test_client.Stream('fake-stream')
        m.get(URL_STREAM_ID.format(s.name), status_code=404)
        assume(bool(s) == False)

def test_stream_len():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')
        count = 8675309
        m.get(URL_STREAM_ID_META.format(s.name), json={ 'events': count })
        assume(len(s) == count)

def test_stream_stats():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')
        payload = { 'events': 3, 'fields': [['some'], ['fields']], 'healthy': True, 'meta': {} }
        m.get(URL_STREAM_ID_META.format(s.name), json=payload)
        assume(s.stats() == payload)

def test_stream_healthy():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')
        m.get(URL_STREAM_ID_META.format(s.name), json={ 'healthy': True })
        assume(s.healthy() == True)

def test_stream_healthy_404():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')
        m.get(URL_STREAM_ID_META.format(s.name), status_code=404)
        assume(s.healthy() == None)

def test_empty_range():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')

        start = datetime(2000,1,1)
        end = datetime(2000,1,2)
        m.get(range_url(s, start, end), text='')

        results = s.range(start, end).df()
        assume(len(results) == 0)

def test_empty_select():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')
        loc = '1234'

        m.post("/query", headers={ 'location': loc })
        m.get("/query/" + loc + "/spans", json={ 'spans': [] })

        results = test_client.select(s.temp > 1000).df()
        assume(len(results) == 0)

def test_stream_values():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')

        when = datetime(2018,1,1,3,55)
        url_when = "2018-01-01T03%3A55%3A00%2B00%3A00"

        m.get(URL_VALUES.format(s.name) + '?at=' + url_when,
            json=[
                {'ts': '2017-12-04T04:00:00Z', 'path': ['temp'], 'value': 37.47, 'id': '123'},
                {'ts': '2017-12-05T04:00:00Z', 'path': ['nested', 'humidity'], 'value': 0.75, 'id': 'abc'},
            ]
            )
        items = s.values(at=when).items()

        assert len(items) == 2
        assert str(items[0][0]) == '(stream "weather"):temp'
        assert str(items[1][0]) == '(stream "weather"):nested.humidity'
        assert items[0][1] == 37.47
        assert items[1][1] == 0.75

def test_create_event():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')

        evt = s.Event(id='55', data={ 'temp': 32 }, ts=datetime(2018,1,1,3,55))
        assume(evt.exists == False)

        m.put(URL_EVENTS_ID.format(s.name, evt.id))
        evt.create()
        assume(evt.exists == True)

def test_read_event():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')

        evt = s.Event(id='55')

        m.get(URL_EVENTS_ID.format(s.name, evt.id),
            headers={
                'location': evt.id,
                'timestamp': '2011-07-21 18:00:00'
                },
            json={
                'temp': 32
            }
            )
        evt.read()
        assume(evt.exists == True)
        assume(evt.json()['event']['temp'] == 32)

def test_delete_event():
    with requests_mock.mock() as m:
        s = test_client.Stream('weather')
        evt = s.Event(id='55', data={ 'temp': 32 }, ts=datetime(2018,1,1,3,55), saved=True)

        m.delete(URL_EVENTS_ID.format(s.name, evt.id))
        evt.delete()
        assume(evt.exists == False)

@given(text())
@example(None)
def test_delete_sid_typechecks(sid):
    with pytest.raises(TypeError):
        test_client.delete(sid, "")

@given(text(min_size=1))
def test_delete_with_eid(eid):
    s = stream("foo")

    def mock_encodings(m, quoter):
        try:
            m.delete(URL_EVENTS_ID.format(s.name, quoter(eid)))
        except:
            pass

    with requests_mock.mock() as m:
        mock_encodings(m, lambda x: x)
        mock_encodings(m, quote)
        mock_encodings(m, lambda x: quote(x.encode('utf-8', 'ignore')))

        assume(test_client.delete(s, eid) == None)





import pytest
import re
import datetime
from datetime import timedelta, datetime

from hypothesis            import given, assume, example, settings
from hypothesis.strategies import text, dictionaries, booleans, integers, floats, lists, dates, datetimes, times, one_of
from sentenai              import hql, V
from sentenai.api.stream   import Stream
from sentenai.historiQL    import StreamPath, Cond, Serial, Switch, And, Or, Select, Par, ast
from shapely.geometry      import Point, Polygon
# Hypothesis Strategies
# ========================

def all_types():
    # FIXME: add times(), having trouble with datetimes in py2 via virtualtime
    return one_of(text(), booleans(), integers(), floats(), dates())

def numeric():
    return one_of(integers(), floats())

def all_multitypes():
    # FIXME: add lists of datetime types
    return one_of(
        lists(text(), min_size=1),
        lists(booleans(), min_size=1),
        lists(integers(), min_size=1),
        lists(floats(), min_size=1))
        #lists(dates(), min_size=1),
        #lists(datetimes(), min_size=1))


# Conditional Checks
# ========================

def stream(name, *filters):
    return Stream(None, name, {}, None, True, *filters)

def assume_parsable(query):
    assume(type(query) == Cond)
    assume(type(ast(query)) == str)

def assert_parsable(query):
    assert (type(query) == Cond)
    assert (type(ast(query)) == str)


def check_syntax_with_type(query, is_multitype=False):
    s = stream("")

    assume_parsable(s.foo.bar.baz.qux == query)
    assume_parsable(s.foo.bar.baz.qux >  query)
    assume_parsable(s.foo.bar.baz.qux >= query)
    assume_parsable(s.foo.bar.baz.qux <  query)
    assume_parsable(s.foo.bar.baz.qux <= query)
    assume_parsable(s.foo.bar.baz.qux != query)

    with pytest.raises(TypeError):
        s.foo >> query

    with pytest.raises(TypeError):
        s.foo << query

    if is_multitype:
        pass
        #assume_parsable(s.foo in query)
        #with pytest.raises(TypeError):
        #    s.foo in query
    else:
        with pytest.raises(TypeError):
            s.foo in query


# Conditional Tests
# ========================
with settings(max_examples=1000, min_satisfying_examples=500):
    @given(all_types())
    def test_conditionals(query):
        check_syntax_with_type(query)

    @given(all_multitypes())
    @example([])
    def test_multibools_conditionals(query):
        check_syntax_with_type(query, is_multitype=True)

    def test_text_wildcard():
        s = stream("")
        assume_parsable(s.foo == "test*")

    def test_text_regex():
        s = stream("")
        assume_parsable(s.foo == r'')

    @given(numeric(), numeric(), numeric())
    def test_incircle(kmx, kmy, rad):
        s = stream("")
        incircle = hql.within_distance(rad, Point(kmx, kmy))
        assume_parsable(s.foo == incircle)

def test_inpoly():
    s = stream("")
    inpoly = hql.inside_region(Polygon([(0, 0), (1, 1), (1, 0)]))
    assert_parsable(s.foo == inpoly)


# Conditional Tests
# ========================

with settings(max_examples=1000, min_satisfying_examples=500):

    @given(all_types())
    def test_simple_conds(query):
        s1 = stream("1")
        s2 = stream("2")

        cond1 = s1.foo == query
        cond2 = s2.bar == query

        assume(type(cond1) is Cond)
        assume(type(cond2) is Cond)


    # @given(all_types(), all_types())
    # def test_switch_construction(query1, query2):
    #     spike = event(V.foo == query1) >> event(V.bar < query2) >> event(V.baz != query1)
    #     assume(isinstance(spike, Switch))

    #     s = stream("1")
    #     assume(isinstance(s(spike), Switch))


    @given(all_types())
    def test_overloaded_ops(query):
        s1 = stream("1")
        s2 = stream("2")

        cond1 = s2.foo == query
        cond2 = s1.foo == query

        assume(type(cond1 & cond2) == And)
        assume(type(cond1 | cond2) == Or)
        # assume(type(span1 >> span2) == Serial)


    # @given(all_types(), all_types())
    # def test_serial_construction(query1, query2):
    #     s1 = stream("1")
    #     s2 = stream("2")
    #     evt1 = event(V.foo == query1)
    #     evt2 = event(V.foo == query2)

    #     srl = span(s1(evt1)) \
    #          .then(s2(evt2), within=delta(days=5))

    #     assume(type(srl) is Serial)


    @given(all_types(), all_types())
    def test_par_construction(query1, query2):
        s1 = stream("1")
        s2 = stream("2")

        assume( isinstance(hql.Any(s1.foo == query1, s2.foo == query2), Par) )
        assume( isinstance(hql.All(s1.foo == query1, s2.foo == query2), Par) )

        # FIXME: this is not working as expected
        #assume(type(all_of(s1(evt1), s2(evt2))()) == dict)


    # @given(all_types(), all_types())
    # def test_select_construction(query1, query2):
    #     s1 = stream("1")
    #     s2 = stream("2")
    #     evt1 = event(V.foo == query1)
    #     evt2 = event(V.foo == query2)

    #     sel0 = select() \
    #         .span(s1(evt1))
    #     assume(isinstance(sel0, Select))

    #     sel1 = select() \
    #         .span(s1(evt1) >> s2(evt2))
    #     assume(isinstance(sel1, Select))

    #     sel3 = select() \
    #         .span(all_of(s1(evt1), s2(evt2)))
    #     assume(isinstance(sel3, Select))

    #     sel4 = select() \
    #         .span((s1(evt1), s2(evt2))) \
    #         .then(s2(evt1))

    #     assume(isinstance(sel4, Select))


# def test_select_with_bounds():
#     s = stream("1")
#     # FIXME: using this as a span returns a TypeError
#     evt = event(V.foo == 1)

#     sel = select(start=datetime.now(), end=datetime.now()) \
#             .span(s.evt == 1) \
#             .then(s.evt == 2)
#             #.span(s(evt) >> s(evt)) \
#             #.then(s(evt))

#     assume(isinstance(sel, Select))
#     assume(type(sel()) == dict)

#     sel = select(end=datetime.now()) \
#             .span(s.evt == 1) \
#             .then(s.evt == 2)
#             #.span(s(evt) >> s(evt)) \
#             #.then(s(evt))

#     assume(isinstance(sel, Select))
#     assume(type(sel()) == dict)

#     sel = select(start=datetime.now()) \
#             .span(s.evt == 1) \
#             .then(s.evt == 2)
#             #.span(s(evt) >> s(evt)) \
#             #.then(s(evt))

#     assume(isinstance(sel, Select))
#     assume(type(sel()) == dict)





import pytest

from hypothesis            import given, assume, example
from hypothesis.strategies import text, dictionaries
from sentenai.api.stream   import Stream
from sentenai.historiQL    import StreamPath
from sentenai.hql          import V

def stream(name, *filters):
    return Stream(None, name, {}, None, True, *filters)

@given(text())
def test_named_streams(name):
    s = stream(name)
    assume(s.name == name)


@given(text())
def test_stream_equality(name):
    s1 = stream(name)
    s2 = stream(name)
    s3 = stream(name + " lies")

    assume(s1 == s2)
    assume(not s1 is s2)
    assume(not (s1 == s3))

@given(text())
def test_filtered_streams(name):
    s1 = stream(name)
    s2 = stream(name)
    s3 = stream(name + " lies")

    assume(s1.filtered(V.temp > 32) == stream(name, V.temp > 32))
    assume(s1.filtered(V.temp > 100) != stream(name, V.temp < 0))
    assume(s3.filtered(V.temp > 32) != stream(name, V.temp > 32))

    s4 = s1.filtered(V.temp > 1000)
    assume(s4.filtered(V.temp > 32, replace=True) == stream(name, V.temp > 32))

@given(text(min_size=1), dictionaries(text(min_size=1), text()))
def test_stream_properties(name, meta):
    s = stream(name, meta)
    assume(s.name == name)
    s._set(name+" new")
    assume(s.name == name + " new")
    assume(s.meta == meta)

@given(text(min_size=1), dictionaries(text(min_size=1), text()))
def test_happy_stream_getitems(name, meta):
    s = stream(name, meta)
    assume(s.meta == meta)
    assume(s.name == name)

@given(text(min_size=1))
def test_stream_to_string(name):
    """ TODO: there are some behaviors around str() which could be tested """
    to_str = str(stream(name))
    pass

@given(text())
@example(path="name")
@example(path="meta")
def test_stream_path_conversions(path):
    s = stream("")
    assume(type(s[path]) == StreamPath)

@given(text())
def test_stream_calls(path):
    s = stream("")
    with pytest.raises(TypeError):
       s(path)



import pytest
import sentenai

def pytest_addoption(parser):
    parser.addoption("--host", action="store", default="http://localhost:8000", help="add host for testing")
    parser.addoption("--auth", action="store", default="", help="auth: add auth key for testing")


@pytest.fixture(scope="module")
def client(request):
    host = request.config.getoption("--host")
    auth = request.config.getoption("--auth")
    return sentenai.Client(auth_key=auth, host=host)
